/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>

// this contains the font interface
#include "gfx.h"
#include "font.h"
#include "womap.h"

static int imax2(int a,int b) { return a>b?a:b; }

/***

  we have only one font, and that is OK.

 ***/

static stbtt_fontinfo    fontinfo;
static float             font_scale_x, font_scale_y;
       int               fontW, fontH;
static int               fontA;
static womap_t          *glyphs;

static void load_mem_font()
{
  const unsigned char *font_data;
  gfx_get_fontfile(4, &font_data, NULL); 
  stbtt_InitFont(&fontinfo, font_data,
                 stbtt_GetFontOffsetForIndex(font_data,0));
}

static void fit_in_rect
  (int rootw, int rooth, int *pixelw, int *pixelh,
   int fontw, int fonth)
{
  float faspect , saspect;
  int pw, ph;

  faspect= (float) fonth / fontw;

  pw= rootw / 80;
  ph= rooth / 25;

  saspect= (float) ph / pw;
  if (saspect > faspect)
  {   
    // screen is taller than necessary, we will base the size on pw.
    ph = pw * faspect;
  }
  else
  {
    // screen is wider than necessary, we will base the size on ph
    pw = ph / faspect;
  }

  *pixelw= pw;
  *pixelh= ph;
}

static void get_char_metrics(unsigned int chr, int *cW, int *cA, int *cD)
{
  int x0, y0, x1, y1;
  stbtt_GetCodepointBox(&fontinfo, chr, &x0, &y0, &x1, &y1);
  *cW= x1-x0;
  *cA= y1;
  *cD= -y0;
}

static void get_font_metrics(int *Rfw, int *Rfa, int *Rfd)
{
  char *metrics_chars= "MWym{}|Q^z#$`gj@";
  int i;
  int fa, fw, fd;
  int ca, cw, cd;

  fw= fa= fd= 0;
  for(i=0;metrics_chars[i];i++)
  {
    get_char_metrics((unsigned char) metrics_chars[i], &cw, &ca, &cd);
    fw= imax2( fw, cw );
    fa= imax2( fa, ca );
    fd= imax2( fd, cd );
  }
  *Rfw= fw;
  *Rfa= fa;
  *Rfd= fd;
}

static void make_light_boxdraw()
{
/**
  Generally speaking, box drawing characters are often problematic with
  truetype fonts. In the best case, everything gets aligned, but there
  are gaps at the edges of horizontal/vertical lines, which is OK but
  doesn't look as good. Therefore, I'm going to patch the font at runtime,
  using the final values for the font width and height.

  Now, there are a bunch of these. I will make the light ones first, and
  then see if there is a need to do the heavier/dashed stuff.

  Light:
    2500        horizontal                 2502        vertical
    250C        upper left corner          2510        upper right corner
    2514        lower left corner          2518        lower right corner
    251C        tee pointing right         2524        tee pointing left
    252C        tee pointing down          2534        tee pointing up
    253C        cross

 These can be rendered as a combination of the following line segments:

                2
                |
            1 =   =  4
                |
                8

 So, here is the table.
**/
  static const int light_boxdraw[]= {
  0x2500,  5, 0x2502, 10, 0x250C, 12, 0x2510, 9, 0x2514, 6, 0x2518, 3,
  0x251C, 14, 0x2524, 11, 0x252C, 13, 0x2534, 7, 0x253C, 15, 0, 0 }; 

  int i;
  glyph_t *G;
  unsigned int ucode, segs;
  int Mx, My;
  int sx, ex, sy, ey, x, y;

  Mx= fontW / 2;
  My= fontH / 2;

  for(i=0;light_boxdraw[i];i+=2)
  {
    ucode= light_boxdraw[i];
    segs= light_boxdraw[i+1];

    G= malloc(sizeof(*G));
    G->width= fontW; G->height= fontH; G->ascent= fontA;
    G->data= calloc(1,G->width*G->height);
    womap_insert(glyphs, ucode, G);

    sx= ex= Mx;  sy= ey= My;

    if (segs&1) sx= 0;
    if (segs&4) ex= fontW-1;
    if (segs&2) sy= 0;
    if (segs&8) ey= fontH-1;

    if (sx!=ex) for(x=sx;x<=ex;x++) G->data[x+My*fontW]= 0xff;
    if (sy!=ey) for(y=sy;y<=ey;y++) G->data[Mx+y*fontW]= 0xff;
  }
}

void  init_font(int pointsz,int wextra, float hscale)
{
  int fw, fa, fd, fh;
  int pixelw, pixelh;
  int rootw, rooth;
  int dpi;

  load_mem_font();
  gfx_display_info(&dpi, &rootw, &rooth);
  get_font_metrics(&fw, &fa, &fd);
  fa *= hscale;
  fd *= hscale;
  fh= fa + fd;

  pixelh= pointsz * dpi / 72.0;
  pixelw= pixelh * fw / fh;

  if (pixelh*25 > rooth || pixelw*80 > rootw)
    fit_in_rect(rootw, rooth, &pixelw, &pixelh, fw, fh);

  font_scale_x= (float) pixelw / fw;
  font_scale_y= (float) pixelh / fh;
 
  fontA= fa * font_scale_y;
  fontH= fh * font_scale_y;
  fontW= fw * font_scale_x + wextra;

  glyphs= womap_new();
  make_light_boxdraw();
}


glyph_t *get_glyph(unsigned int ch)
{
   glyph_t *G;

   unsigned char *src, *osrc;
   int w, h, xoff, yoff;
   int x,y, dx, dy, cw;

   G= womap_find(glyphs, ch);
   if (G) return G;

   G= calloc(1,sizeof(*G));
   G->width= fontW;
   G->height= fontH;
   G->data= calloc(1, fontW*fontH);
   womap_insert(glyphs, ch, G);

   osrc= src= stbtt_GetCodepointBitmap(&fontinfo, font_scale_x, font_scale_y,
                                       ch, &w, &h, &xoff, &yoff);

   /** here, xoff and yoff are the offsets from which the drawing should
       start. the top left corner of the bitmap should be drawn at:
 
       current_point + (xoff,yoff)
    **/


   if (w>fontW) { cw= fontW; src+= (w-fontW)/2; }
           else { cw= w; }
   if (xoff<0)  { dx= 0; }
          else if (xoff+cw>=fontW) { dx=0; } //dx= (fontW-cw)/2; }
                              else { dx= xoff; }

   if (yoff<0)   // normal case
   {
     if (yoff+fontA>=0)
     {
       dy= yoff + fontA;
     }
     else
     {
       dy= 0;
       src += (-yoff-fontA)*w; 
       h -= (-yoff-fontA);
     }
   }
   else
   {
     if (yoff+h>fontH)
     {
       h= fontH-yoff;
       if (h<0) h= 0;
     }
     dy= yoff + fontA;
   }

   for(y=0;y<h;y++)
     for(x=0;x<cw;x++)
       G->data[(x+dx)+(y+dy)*fontW]= src[x+y*w];

   stbtt_FreeBitmap(osrc, NULL);

   return G;
}

